package com.supermap.sample.serviceprovider;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.apache.commons.io.IOUtils;
import com.supermap.services.components.commontypes.Layer;
import com.supermap.services.components.commontypes.MapParameter;
import com.supermap.services.components.commontypes.OutputFormat;
import com.supermap.services.components.commontypes.Point2D;
import com.supermap.services.components.commontypes.PrjCoordSys;
import com.supermap.services.components.commontypes.Rectangle;
import com.supermap.services.components.commontypes.Rectangle2D;
import com.supermap.services.components.commontypes.RectifyType;
import com.supermap.services.components.commontypes.Unit;
import com.supermap.services.providers.RemoteTileProviderBase;
import com.supermap.services.util.TileTool;
import com.supermap.services.util.Tool;

/**
 * <p>
 * 谷歌地图  地图服务提供者
 * </p>
 * <p>
 * GoogleMapsMapProvider 提供了聚合第三方的Google地图服务的能力，并封装了与 SuperMap iServer 地图相关的 GIS
 * 功能。
 * </p>
 * <p>
 * Google Map  Map service provider
 * </p>
 * <p>
 * GoogleMapsMapProvider provides the ability that integrates third-party Google map services, and encapsulates the GIS functions related with SuperMap iServer maps
 * </p>
 */
public class GoogleMapsMapProvider extends RemoteTileProviderBase {
    private static final String GOOGLEMAPSADRESSTEMPLATE = "http://mt1.google.cn/vt/hl=zh-CN&x=%d&y=%d&z=%d";
    private static final String MAPNAME = "googlemaps";
    private static final long EARTH_RADIUS = 6378137;
    private static final int LEVEL_COUNT = 18;
    private static final double GOOGLE_DPI = 96.0;
    private static final double RESOLUTION_PRECISION = 1E-6;
    private double[] resolutions;

    /**
     * <p>
     * 获取谷歌地图瓦片URL
     * </p>
     * Gets the google mao tiles URL
     * @param tileParam
     */
    @Override
    protected String getTileImageUrl(TileImagePrameter tileImagePrameter) {
        int z = getLevelByResolution(tileImagePrameter.resolution);
        String googleImageUrl = String.format(GOOGLEMAPSADRESSTEMPLATE, tileImagePrameter.x, tileImagePrameter.y, z);
        return googleImageUrl;
    }

    @Override
    protected boolean cacheEnabled() {
        return false;
    }

    @Override
    protected Point2D getOrigin(String arg0) {
        return new Point2D(-20037508.3427892, 20037508.3427892);
    }

    /**
    *
    * <p>
    * 返回指定地图支持的图片格式数组
    * </p>
    * Return picture format array supported by the specified map
    * @param mapName
    * @return
    * @since 7.1.0
    */
    @Override
    public OutputFormat[] getSupportImageFormat(String arg0) {
        return new OutputFormat[0];
    }

    @Override
    public double[] getSupportResolutions(String arg0) {
        if (resolutions == null) {
            /**
             * 第0级分辨率 = 地球赤道周长/256个像素
             * 0 level resolution = Circumference of the earth's equator/256 pixels
             */
            double resolution = 2 * Math.PI * EARTH_RADIUS / 256.0;
            resolutions = new double[LEVEL_COUNT];
            for (int i = 0; i < LEVEL_COUNT; i++) {
                resolutions[i] = resolution;
                resolution /= 2.0;
            }
        }
        return resolutions.clone();
    }

    /**
     * <p>
     * 初始化默认地图状态，返回值为一个映射，键为地图名，值为地图参数对象。
     * </p>
     * <p>
     * Initialize the defaule map status. The return value is a mapping. The key is the map name and the value is the map parameter object.
     * </p>
     * @return 地图状态映射，键为地图名，值为地图参数对象。
     * <p>
     * mapping.The key is the map name, and the value is the map parameter object.
     * </p>
     */
    @Override
    protected Map<String, MapParameter> initDefaultMapParameter() {
        String mapName = MAPNAME;
        MapParameter defaultMapParameter = new MapParameter();
        defaultMapParameter.name = mapName;
        defaultMapParameter.coordUnit = Unit.METER;
        defaultMapParameter.distanceUnit = Unit.METER;
        defaultMapParameter.viewer = new Rectangle(0, 0, 256, 256);
        defaultMapParameter.bounds = new Rectangle2D(-20037508.3427892, -20037508.3427892, 20037508.3427892, 20037508.3427892);
        defaultMapParameter.viewBounds = new Rectangle2D(defaultMapParameter.bounds);
        defaultMapParameter.center = defaultMapParameter.viewBounds.center();
        defaultMapParameter.customEntireBoundsEnabled = false;
        defaultMapParameter.description = "";
        defaultMapParameter.prjCoordSys = new PrjCoordSys();
        defaultMapParameter.prjCoordSys.coordUnit = Unit.METER;
        defaultMapParameter.prjCoordSys.distanceUnit = Unit.METER;
        defaultMapParameter.prjCoordSys.epsgCode = 3857;
        defaultMapParameter.rectifyType = RectifyType.BYCENTERANDMAPSCALE;
        defaultMapParameter.visibleScalesEnabled = true;
        defaultMapParameter.visibleScales = getSupportScales();
        defaultMapParameter.scale = defaultMapParameter.visibleScales[0];
        defaultMapParameter.layers = new ArrayList<Layer>();
        GoogleMapLayer layer = new GoogleMapLayer();
        layer.name = defaultMapParameter.name;
        layer.caption = defaultMapParameter.name;
        layer.bounds = new Rectangle2D(defaultMapParameter.bounds);
        layer.description = "Google Maps Layer";
        layer.visible = true;
        defaultMapParameter.layers.add(layer);
        Map<String, MapParameter> result = new HashMap<String, MapParameter>();
        result.put(mapName, defaultMapParameter);
        return result;
    }

    /**
     * 初始化支持的地图列表
     *<p>
     *Initialize the supported map list
     *</p>
     * @return 地图名列表
     * <p>
     * map name list
     * </p>
     */
    @Override
    protected List<String> initSupportedMapNames() {
        List<String> names = new ArrayList<String>();
        names.add(MAPNAME);
        return names;
    }

    private int getLevelByResolution(double resolution) {
        resolutions = getSupportResolutions(MAPNAME);
        for (int i = 0; i < resolutions.length; i++) {
            if (Tool.equal(resolution, resolutions[i], RESOLUTION_PRECISION)) {
                return i;
            }
        }
        return -1;
    }

    private double[] getSupportScales() {
        double[] resolutions = getSupportResolutions(MAPNAME);
        double[] scales = new double[LEVEL_COUNT];
        for (int i = 0; i < LEVEL_COUNT; i++) {
            scales[i] = TileTool.resolutionToScale(resolutions[i], GOOGLE_DPI, Unit.METER);
        }
        return scales;
    }

    /**
     * <p>
     * 根据参数向远端发送出图请求
     * </p>
     * <p>
     * Send a request of outputting maps to the remote according to the parameters
     * </p>
     */
    @Override
    protected Map<String, Object> sendRequestByHttpURLConnection(String paramString1, String paramString2, InputStream paramInputStream,
            Map<String, String> paramMap, int paramInt) {
        InputStream localInputStream = null;
        HttpURLConnection localHttpURLConnection = null;
        OutputStream localOutputStream = null;
        int i = paramInt;
        HashMap<String, Object> localHashMap = new HashMap<String, Object>();
        try {
            URL localURL = new URL(paramString1);
            localHttpURLConnection = (HttpURLConnection) localURL.openConnection();
            localHttpURLConnection.setRequestMethod(paramString2);
            if ((paramString2.equals("POST")) || (paramString2.equals("PUT"))) {
                localHttpURLConnection.setDoOutput(true);
            }
            localHttpURLConnection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded; charset=UTF-8");
            localHttpURLConnection.setRequestProperty("User-Agent",
                    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.116 Safari/537.36");
            localHttpURLConnection.setRequestProperty("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
            localHttpURLConnection.setRequestProperty("Accept-Encoding", "gzip,deflate,sdch");
            localHttpURLConnection.setRequestProperty("Accept-Language", "zh-CN,zh;q=0.8");
            if (paramMap != null) {
                Iterator<?> localIterator = paramMap.entrySet().iterator();
                while (localIterator.hasNext()) {
                    @SuppressWarnings("rawtypes")
                    Map.Entry localEntry = (Map.Entry) localIterator.next();
                    localHttpURLConnection.setRequestProperty((String) localEntry.getKey(), (String) localEntry.getValue());
                }
            }
            localHttpURLConnection.setConnectTimeout(40000);
            localHttpURLConnection.setReadTimeout(20000);
            if ((paramInputStream != null) && (localHttpURLConnection.getDoOutput())) {
                localOutputStream = localHttpURLConnection.getOutputStream();
                IOUtils.copy(paramInputStream, localOutputStream);
                localOutputStream.flush();
                localOutputStream.close();
                paramInputStream.close();
            }
            int responseCode = localHttpURLConnection.getResponseCode();
            localHashMap.put("status", Integer.valueOf(responseCode));
            if (responseCode < 400) {
                localInputStream = localHttpURLConnection.getInputStream();
            } else {
                logger.info(message.getMessage("TiledMapProvider.sendRequestByHttpURLConnection.RequestError", new String[] { paramString1, String.valueOf(responseCode) }));
            }
            if (localInputStream != null)
                localHashMap.put("ENTITY", localInputStream);
        } catch (IOException localIOException) {
            if (i < 5) {
                logger.debug(message.getMessage("URL_REQUEST_EXCEPTION", new String[] { paramString1, localIOException.getMessage(), String.valueOf(i) }));
                return sendRequestByHttpURLConnection(paramString1, paramString2, paramInputStream, ++i);
            }
            logger.debug(message.getMessage("URL_REQUEST_FAILED", new String[] { paramString1, String.valueOf(i) }));
        }
        return localHashMap;
    }
}
